import click
import pandas as pd
from sqlalchemy import create_engine, text
from sqlalchemy.exc import SQLAlchemyError

from hckr.cli.config import common_config_options
from hckr.utils.DataUtils import print_df_as_table
from hckr.utils.DbUtils import get_db_url
from hckr.utils.MessageUtils import PError, PInfo


@click.group(
    help="Database commands",
    context_settings={"help_option_names": ["-h", "--help"]},
)
def db():
    pass


@db.command()
@common_config_options
@click.argument("query")
@click.option(
    "-nr",
    "--num-rows",
    default=10,
    help="Number of rows to show.",
    required=False,
)
@click.option(
    "-nc",
    "--num-cols",
    default=10,
    help="Number of cols to show.",
    required=False,
)
@click.pass_context
def query(ctx, config, query, num_rows=None, num_cols=None):
    """
    This command executes a SQL query on your configured database and show you result in a table format ( in
    ``SELECT/SHOW/DESC`` queries )

    **Example Usage**:

    * Running a simple query on your configured database

    .. code-block:: shell

        $ hckr db query "select 1 as key, 'one' as value" -c testdb_sqlite

    .. table:: Output
       :widths: auto

       =====    =========
       key      value
       =====    =========
       1        one
       =====    =========


    .. note::
       here, we have configured ``Sqlite`` database with in a config 'testdb_sqlite'
       using :ref:`Configuring your databases <configuration>`


    * Additionally, we can also limit number of records and columns returned by the query using ``-nr/--num-rows`` and ``-nc/--num-cols`` options

    .. code-block:: shell

        $ hckr db query "select 1 as key, 'one' as value" -c testdb_sqlite -nr 1 -nc 1

    **Command Reference**:
    """
    db_url = get_db_url(section=config)
    if not db_url:
        PError("Database credentials are not properly configured.")

    query = query.strip()
    engine = create_engine(db_url)
    try:
        with engine.connect() as connection:
            # Normalize and determine the type of query
            normalized_query = query.lower()
            is_data_returning_query = normalized_query.startswith(
                ("select", "desc", "describe", "show", "explain")
            )
            is_ddl_query = normalized_query.startswith(
                ("create", "alter", "drop", "truncate")
            )

            if is_data_returning_query:
                # Execute and fetch results for queries that return data
                df = pd.read_sql_query(text(query), connection)

                # Optionally limit rows and columns if specified
                if num_rows is not None:
                    df = df.head(num_rows)
                if num_cols is not None:
                    df = df.iloc[:, :num_cols]

                print_df_as_table(df, title=query)
                return df
            else:
                # Execute DDL or non-data-returning DML queries
                with connection.begin():  # this will automatically commit at the end
                    result = connection.execute(text(query))
                if is_ddl_query:
                    PInfo(query, "Success")
                else:
                    PInfo(query, f"[Success] Rows affected: {result.rowcount}")
    except SQLAlchemyError as e:
        PError(f"Error executing query: {e}")



import click

from ..utils.MessageUtils import PError, PSuccess
from ..utils.config.ConfigUtils import (
    init_config,
    DEFAULT_CONFIG,
    configMessage,
    list_config,
    set_config_value,
    get_config_value,
)


@click.group(
    help="Config commands",
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.pass_context
def config(ctx):
    """
    Defines a command group for managing application configurations. This group includes commands to set, get, show, and initialize configuration values.
    """
    pass


def common_config_options(func):
    func = click.option(
        "-c",
        "--config",
        help="Config instance, default: DEFAULT",
        default=DEFAULT_CONFIG,
    )(func)
    return func


@config.command()
@common_config_options
@click.argument("key")
@click.argument("value")
def set(config, key, value):
    """
    This command adds a new entry to the config file with key and value

    **Example Usage**:

    * Setting a value inside DEFAULT config

    .. code-block:: shell

        $ hckr config set database_host 127.0.0.1

    * Similarly, we can also set a value in specific configuration, configuration will be created if not exists

    .. code-block:: shell

        $ hckr config set database_host 127.0.0.1 --config MY_DATABASE

    **Command Reference**:
    """

    configMessage(config)
    set_config_value(config, key, value)
    PSuccess(f"[{config}] {key} <- {value}")


@config.command()
@common_config_options
@click.argument("key")
def get(config, key):
    """
    This command returns value for a key in a configuration

    **Example Usage**:

    * Getting a value for key in DEFAULT config

    Note - DEFAULT config is parent of all other configurations and others will inherit its values if not overridden

    .. code-block:: shell

        $ hckr config get database_host

    * Similarly, we can also get a value in specific configuration

    .. code-block:: shell

        $ hckr config get database_host --config MY_DATABASE

    **Command Reference**:
    """

    configMessage(config)
    try:
        value = get_config_value(config, key)
        PSuccess(f"[{config}] {key} = {value}")
    except ValueError as e:
        PError(f"{e}")


@config.command()
@common_config_options
@click.option(
    "-a",
    "--all",
    default=False,
    is_flag=True,
    help="Whether to show all configs (default: False)",
)
def show(config, all):
    """
    This command show list of all keys available in given configuration,
    we can also see values in all configurations by providing -a/--all flag

    **Example Usage**:

    * Getting values for keys in DEFAULT config

    Note - DEFAULT config is parent of all other configurations and others will inherit its values if not overridden

    .. code-block:: shell

        $ hckr config show

    * Similarly, we can also get all values in a specific configuration using -c/--config flag

    .. code-block:: shell

        $ hckr config show -c MY_DATABASE

    * Additionally, we can also see all configurations using -a/--all flag

    .. code-block:: shell

        $ hckr config show --all

    **Command Reference**:
    """
    list_config(config, all)


@config.command()
@click.option(
    "-o",
    "--overwrite",
    default=False,
    is_flag=True,
    help="Whether to delete and recreate .hckrcfg file (default: False)",
)
def init(overwrite):
    """
     This command Initializes the configuration for the application,
    we can also use this option to overwrite existing config and reinitialize.

    **Example Usage**:

    * Initialising config file .hckrcfg with default settings

    .. code-block:: shell

        $ hckr config init

    * Similarly, we can also delete existing file and recreate using -o/--overwrite flag

    .. code-block:: shell

        $ hckr config init --overwrite

    **Command Reference**:
    """
    init_config(overwrite)



# SPDX-FileCopyrightText: 2024-present Ashish Patel <ashishpatel0720@gmail.com>
#
# SPDX-License-Identifier: MIT
from pathlib import Path

import pytest
from click.testing import CliRunner

from hckr.utils.config.ConfigUtils import init_config

current_directory = Path(__file__).parent
SQLITE_DB = current_directory / "cli" / "resources" / "db" / "test_db.sqlite"


@pytest.fixture(scope="package")
def local_sqlite_db():
    return SQLITE_DB


@pytest.fixture(scope="package")
def cli_runner():
    init_config(overwrite=True)  # recreate the config file
    return CliRunner()


@pytest.fixture(scope="package")
def postgres_options():
    return [
        "--config-name",
        "testdb_postgres",
        "--database-type",
        "1",
        "--user",
        "user",
        "--password",
        "password",
        "--host",
        "host.postgres",
        "--port",
        "11143",
        "--database-name",
        "defaultdb",
    ]


@pytest.fixture(scope="package")
def mysql_options():
    return [
        "--config-name",
        "testdb_mysql",
        "--database-type",
        "2",
        "--user",
        "user",
        "--password",
        "password",
        "--host",
        "host",
        "--port",
        "123",
        "--database-name",
        "defaultdb",
    ]


@pytest.fixture(scope="package")
def snowflake_options():
    return [
        "--config-name",
        "testdb_snowflake",
        "--database-type",
        "4",
        "--user",
        "user",
        "--password",
        "password",
        "--database-name",
        "database",
        "--schema",
        "PUBLIC",
        "--account",
        "account-id",
        "--warehouse",
        "COMPUTE_WH",
        "--role",
        "ACCOUNTADMIN",
    ]


@pytest.fixture(scope="package")
def sqlite_options():
    return [
        "--config-name",
        "testdb_sqlite",
        "--database-type",
        "3",
        "--database-name",
        f"{SQLITE_DB}",
    ]



from enum import Enum
from pathlib import Path

config_path = Path.home() / ".hckrcfg"
DEFAULT_CONFIG = "DEFAULT"


class DBType(str, Enum):
    PostgreSQL = ("PostgreSQL",)
    MySQL = ("MySQL",)
    SQLite = ("SQLite",)
    Snowflake = ("Snowflake",)

    def __str__(self):
        return self.value


class ConfigType(str, Enum):
    DATABASE = ("database",)

    def __str__(self):
        return self.value


db_type_mapping = {
    "1": DBType.PostgreSQL,
    "2": DBType.MySQL,
    "3": DBType.SQLite,
    "4": DBType.Snowflake,
}

# =============== CONFIG CONSTANTS =================== #
# COMMON CONFIG Constants
CONFIG_TYPE = "config_type"

# DATABASE CONFIG Constants
DB_TYPE = "type"
DB_HOST = "host"
DB_PORT = "port"
DB_USER = "username"
DB_PASSWORD = "password"

DB_NAME = "database"
DB_SCHEMA = "schema"

# SNOWFLAKE SPECIFIC
DB_ACCOUNT = "account"
DB_ROLE = "role"
DB_WAREHOUSE = "warehouse"



import configparser
import logging

import rich
from rich.panel import Panel

from .Constants import config_path, DEFAULT_CONFIG
from .. import MessageUtils
from ..MessageUtils import PWarn, PSuccess, PInfo, PError
from ...__about__ import __version__


def load_config():
    """Load the INI configuration file."""
    config = configparser.ConfigParser()
    if not config_exists():
        PError(
            f"Config file [magenta]{config_path}[/magenta] doesn't exists or empty,"
            f" Please run init command to create one \n "
            "[bold green]hckr config init"
        )
    config.read(config_path)
    return config


def config_exists() -> bool:
    """
    Check if config file exists and is not empty.
    """
    if not config_path.exists():
        return False
    if config_path.stat().st_size == 0:
        return False
    with config_path.open("r") as file:
        content = file.read().strip()
        if not content:
            return False
    return True


def init_config(overwrite):
    if not config_path.exists():
        config_path.parent.mkdir(parents=True, exist_ok=True)
        config_path.touch(exist_ok=True)
        default_config = {
            DEFAULT_CONFIG: {
                "version": f"{__version__}",
                "config_type": "default",
            },
            "CUSTOM": {
                "key": f"value",
            },
        }
        config = configparser.ConfigParser()
        config.read_dict(default_config)
        with config_path.open("w") as config_file:
            config.write(config_file)
        PSuccess(f"Config file created at {config_path}")
    elif overwrite:
        PInfo(
            f"Config file already exists at {config_path}, [magenta]-o/--overwrite[/magenta] passed \n"
            "[yellow]Deleting existing file"
        )
        config_path.unlink()
        init_config(overwrite=False)
    else:
        PWarn(
            f"Config file already exists at [yellow]{config_path}[/yellow],"
            " please pass [magenta]-o/--overwrite[/magenta] to recreate"
        )


def set_config_value(section, key, value, override=False):
    """
    Sets a configuration value in a configuration file.
    """
    logging.debug(f"Setting [{section}] {key} = {value}")
    config = load_config()
    if not config.has_section(section) and section != DEFAULT_CONFIG:
        PInfo(f"Config \[{section}] doesn't exist, Adding")
        config.add_section(section)

    config.set(section, key, value)
    with config_path.open("w") as config_file:
        config.write(config_file)


def get_config_value(section, key) -> str:
    logging.debug(f"Getting [{section}] {key} ")
    config = load_config()
    if section != DEFAULT_CONFIG and not config.has_section(section):
        raise ValueError(f"Section '{section}' not found in the configuration.")
    if not config.has_option(section, key):
        raise ValueError(f"Key '{key}' not found in section '{section}'.")
    return config.get(section, key)


def show_config(config, section):
    if section == DEFAULT_CONFIG:
        rich.print(
            Panel(
                (
                    "\n".join(
                        [f"{key} = {value}" for key, value in config.items("DEFAULT")]
                    )
                    if config.items("DEFAULT")
                    else "NOTHING FOUND"
                ),
                expand=True,
                title="\[DEFAULT]",
            )
        )
    elif config.has_section(section):
        rich.print(
            Panel(
                (
                    "\n".join(
                        [f"{key} = {value}" for key, value in config.items(section)]
                    )
                    if config.items(section)
                    else "NOTHING FOUND"
                ),
                expand=True,
                title=f"\[{section}]",
            )
        )
    else:
        rich.print(
            Panel(
                f"config {section} not found",
                expand=True,
                title="Error",
            )
        )


def list_config(section, all=False):
    config = load_config()
    if all:
        MessageUtils.info("Listing all config")
        show_config(config, DEFAULT_CONFIG)
        for section in config.sections():
            show_config(config, section)
    else:
        configMessage(section)
        show_config(config, section)


def configMessage(config):
    if config == DEFAULT_CONFIG:
        MessageUtils.info(f"Using default config: [magenta]{DEFAULT_CONFIG}")
    else:
        MessageUtils.info(f"Using config: [magenta]{config}")



import click

from .ConfigUtils import set_config_value
from .Constants import (
    DBType,
    DB_USER,
    DB_PASSWORD,
    DB_HOST,
    DB_PORT,
    DB_ACCOUNT,
    DB_WAREHOUSE,
    DB_ROLE,
    DB_SCHEMA,
)


def configure_host(
    account, config_name, host, port, role, schema, selected_db_type, warehouse
):
    if selected_db_type in [DBType.PostgreSQL, DBType.MySQL]:
        if not host:
            host = click.prompt("Enter the database host (e.g., localhost, 127.0.0.1)")
        if not port:
            port = click.prompt("Enter the database port (e.g., 5432 for PostgreSQL)")

        set_config_value(config_name, DB_HOST, host)
        set_config_value(config_name, DB_PORT, port)

    elif selected_db_type == DBType.Snowflake:
        if not account:
            account = click.prompt("Enter the Snowflake Account Id")
        if not schema:
            schema = click.prompt("Enter the Snowflake schema name")
        if not warehouse:
            warehouse = click.prompt("Enter the Snowflake warehouse name")
        if not role:
            role = click.prompt("Enter the Snowflake role")
        set_config_value(config_name, DB_ACCOUNT, account)
        set_config_value(config_name, DB_WAREHOUSE, warehouse)
        set_config_value(config_name, DB_SCHEMA, schema)
        set_config_value(config_name, DB_ROLE, role)


def configure_creds(config_name, password, selected_db_type, user):
    if selected_db_type != DBType.SQLite:
        if not user:
            user = click.prompt("Enter the database user (e.g., root)")
        if not password:
            password = click.prompt(
                "Enter the database password (input hidden)",
                hide_input=True,
                confirmation_prompt=True,
            )
        # common values
        set_config_value(config_name, DB_USER, user)
        set_config_value(config_name, DB_PASSWORD, password)



from click.testing import CliRunner

from hckr.cli import cli
from hckr.__about__ import __version__


def test_hckr():
    runner = CliRunner()
    result = runner.invoke(cli)
    print(result.output)
    assert result.exit_code == 0
    assert __version__ in result.output
    assert "[OPTIONS] COMMAND [ARGS]..." in result.output
    assert (
        """Options:
  -v, --verbose  Enable verbose output, use -v for INFO and -vv for DEBUG
  --version      Show the version and exit.
  -h, --help     Show this message and exit.

Commands:
  config     Config commands
  configure  easy configurations for other commands (eg.
  cron       cron commands
  crypto     crypto commands
  data       data related commands
  db         Database commands
  hash       hash commands
  info       info commands
  k8s        Kubernetes commands
  net        network commands
  repl       Start an interactive shell."""
        in result.output
    )



import click

from ..utils.MessageUtils import PSuccess
from ..utils.config.ConfigUtils import (
    list_config,
    set_config_value,
)
from ..utils.config.ConfigureUtils import configure_host, configure_creds
from ..utils.config.Constants import (
    CONFIG_TYPE,
    DB_TYPE,
    ConfigType,
    db_type_mapping,
    DB_NAME,
)


@click.group(
    help="easy configurations for other commands (eg. db)",
    context_settings={"help_option_names": ["-h", "--help"]},
)
@click.pass_context
def configure(ctx):
    """
    Defines a command group for configuration-related commands.
    """
    pass


@configure.command("db")
@click.option(
    "--config-name",
    prompt="Enter a name for this database configuration",
    help="Name of the config instance",
)
@click.option(
    "--database-type",
    prompt="Select the type of database (1=PostgreSQL, 2=MySQL, 3=SQLite, 4=Snowflake)",
    type=click.Choice(["1", "2", "3", "4"]),
    help="Database type",
)
@click.option("--host", prompt=False, help="Database host")
@click.option("--port", prompt=False, help="Database port")
@click.option("--user", prompt=False, help="Database user")
@click.option(
    "--password",
    prompt=False,  # we will get this value later if it is not provided
    hide_input=True,
    confirmation_prompt=True,
    help="Database password",
)
@click.option("--database-name", prompt=False, help="Database name")
@click.option("--schema", prompt=False, help="Database schema")
@click.option("--account", prompt=False, help="Snowflake Account Id")
@click.option("--warehouse", prompt=False, help="Snowflake warehouse")
@click.option("--role", prompt=False, help="Snowflake role")
def configure_db(
    config_name,
    database_type,
    host,
    port,
    user,
    password,
    database_name,
    schema,
    account,
    warehouse,
    role,
):
    """
    This command configures database credentials based on the selected database type.

    .. hint::
       We currently support ``Postgres``, ``MySql``, ``Sqlite`` and  ``Snowflake``,
       Please feel free to raise an issue or Pull request if additional databases are needed.

    .. table:: Database Options mapping
       :widths: auto

       =====    =====
       1        Postgres
       2        Mysql
       3        Sqlite
       4        Snowflake
       =====    =====

    **Example Usage**:

    * Setting up your database configuration using Prompt values


    .. code-block:: shell

        $ hckr configure db

    .. note::
       Using this cli will ask for information like database ``host``, ``port``, ``username`` etc. as per the database type selected.

    * Similarly, we can also provide all values using flag

    .. code-block:: shell

        $  hckr configure db --config-name test-config --database-type 3  --database-name mydb.sqlite

    .. note::
       here, we have configured ``Sqlite`` database with a name 'mydb.sqlite' by providing both flags in command line,
       for other database types, we will have to provide required flags accordingly.

    **Command Reference**:
    """

    set_config_value(config_name, CONFIG_TYPE, ConfigType.DATABASE)
    selected_db_type = db_type_mapping[database_type]
    set_config_value(config_name, DB_TYPE, selected_db_type)

    configure_creds(config_name, password, selected_db_type, user)

    if not database_name:
        database_name = click.prompt("Enter the database name")
    set_config_value(config_name, DB_NAME, database_name)

    configure_host(
        account, config_name, host, port, role, schema, selected_db_type, warehouse
    )

    PSuccess(
        f"Database configuration saved successfully in config instance '{config_name}'"
    )
    list_config(config_name)



from hckr.cli.configure import configure_db
from hckr.utils.MessageUtils import _PMsg, PInfo


def test_configure_postgres(cli_runner, postgres_options):
    result = cli_runner.invoke(configure_db, postgres_options)
    print(result.output)
    assert result.exit_code == 0
    assert "Database configuration saved successfully" in result.output
    assert "[testdb_postgres]" in result.output


def test_configure_mysql(cli_runner, mysql_options):
    result = cli_runner.invoke(configure_db, mysql_options)
    print(result.output)
    assert result.exit_code == 0
    assert "Database configuration saved successfully" in result.output
    assert "[testdb_mysql]" in result.output


def test_configure_snowflake(cli_runner, snowflake_options):
    result = cli_runner.invoke(configure_db, snowflake_options)
    print(result.output)
    assert result.exit_code == 0
    assert "Database configuration saved successfully" in result.output
    assert "[testdb_snowflake]" in result.output


def test_configure_sqlite(cli_runner, sqlite_options):
    result = cli_runner.invoke(configure_db, sqlite_options)
    print(result.output)
    assert result.exit_code == 0
    assert "Database configuration saved successfully" in result.output
    assert "[testdb_sqlite]" in result.output


def test_test():
    hi = "ashish"
    PInfo(f"hello {hi}")



import logging
from configparser import NoOptionError

from hckr.utils.MessageUtils import PError
from hckr.utils.config import ConfigUtils
from hckr.utils.config.Constants import (
    ConfigType,
    DBType,
    DB_NAME,
    DB_USER,
    CONFIG_TYPE,
    DB_TYPE,
    DB_PASSWORD,
    DB_HOST,
    DB_PORT,
    DB_ACCOUNT,
    DB_WAREHOUSE,
    DB_ROLE,
    DB_SCHEMA,
)


def get_db_url(section):
    """
    This function retrieves a database URL based on the given configuration section.
    :param section: The name of the configuration section to retrieve the database URL from.
    :return: The database URL.
    """
    config = ConfigUtils.load_config()
    try:
        config_type = config.get(section, CONFIG_TYPE)
        if config_type != ConfigType.DATABASE:
            PError(
                f"The configuration {section} is not database type\n"
                " Please use [magenta]hckr configure db[/magenta] to configure database."
            )
        db_type = config.get(section, DB_TYPE)

        if db_type == DBType.SQLite:
            database_name = config.get(section, DB_NAME)
            return f"sqlite:///{database_name}"

        elif db_type in [DBType.PostgreSQL, DBType.MySQL]:
            return _get_jdbc_url(config, section, db_type)

        elif db_type == DBType.Snowflake:
            return _get_snowflake_url(config, section)

    except NoOptionError as e:
        PError(f"Config {section} is not configured correctly\n {e}")


def _get_jdbc_url(config, section, db_type):
    user = config.get(section, DB_USER)
    password = config.get(section, DB_PASSWORD)
    host = config.get(section, DB_HOST)
    port = config.get(section, DB_PORT)
    database_name = config.get(section, DB_NAME)
    if db_type == DBType.PostgreSQL:
        return f"postgresql://{user}:{password}@{host}:{port}/{database_name}"
    elif db_type == DBType.MySQL:
        return f"mysql+pymysql://{user}:{password}@{host}:{port}/{database_name}"
    else:
        logging.debug(f"Invalid db_type: {db_type} in get_jdbc_url()")
        PError("Error occured while creating JDBC Url")


def _get_snowflake_url(config, section):
    user = config.get(section, DB_USER)
    password = config.get(section, DB_PASSWORD)
    account = config.get(section, DB_ACCOUNT)
    warehouse = config.get(section, DB_WAREHOUSE)
    role = config.get(section, DB_ROLE)
    database_name = config.get(section, DB_NAME)
    schema = config.get(section, DB_SCHEMA)
    return (
        f"snowflake://{user}:{password}@{account}/{database_name}/{schema}"
        f"?warehouse={warehouse}&role={role}"
    )



from click.testing import CliRunner

from hckr.cli.db import query
from hckr.cli.configure import configure_db


def _run_query_and_assert(cli_runner, sql_query, value_assert=None):
    result = cli_runner.invoke(query, [sql_query, "-c", "testdb_sqlite"])
    print(result.output)
    if value_assert:
        assert value_assert in result.output


def test_db_query_sqlite(cli_runner, sqlite_options):
    # # configuring sqlite for testing
    result = cli_runner.invoke(configure_db, sqlite_options)
    assert result.exit_code == 0
    assert "Database configuration saved successfully" in result.output
    assert "[testdb_sqlite]" in result.output

    _run_query_and_assert(cli_runner, "drop table if exists users")

    _run_query_and_assert(
        cli_runner,
        """
                              CREATE TABLE users (
                              id INTEGER PRIMARY KEY,
                              name TEXT NOT NULL,
                              email TEXT NOT NULL UNIQUE
                              );
                              """,
    )

    _run_query_and_assert(
        cli_runner,
        "INSERT INTO users (name, email) VALUES ('Alice', 'alice@example.com');",
    )

    _run_query_and_assert(cli_runner, "select * from users;")


# NEGATIVE USE CASES


def test_db_query_missing_or_invalid_config():
    runner = CliRunner()
    result = runner.invoke(query, ["select 1"])
    print(result.output)
    assert result.exit_code != 0
    assert "The configuration DEFAULT is not database type" in result.output


def test_db_query_missing_query():
    runner = CliRunner()
    result = runner.invoke(query)
    print(result.output)
    assert result.exit_code != 0
    assert "Error: Missing argument 'QUERY'" in result.output



import random
import string

from click.testing import CliRunner

from hckr.cli.config import set, get, show


def _get_random_string(length):
    letters = string.ascii_lowercase
    result_str = "".join(random.choice(letters) for i in range(length))
    return result_str


# CONFIG GET AND SET
def test_config_get_set_default(cli_runner):
    _key = f"key_{_get_random_string(5)}"
    _value = f"value_{_get_random_string(5)}"
    result = cli_runner.invoke(set, [_key, _value])
    assert result.exit_code == 0
    assert f"[DEFAULT] {_key} <- {_value}" in result.output

    # testing get
    result = cli_runner.invoke(get, [_key])
    assert result.exit_code == 0
    assert f"[DEFAULT] {_key} = {_value}" in result.output


def test_config_get_set_custom_config(cli_runner):
    _key = f"key_{_get_random_string(5)}"
    _value = f"value_{_get_random_string(5)}"
    _CONFIG = "CUSTOM"
    result = cli_runner.invoke(set, ["--config", _CONFIG, _key, _value])
    assert result.exit_code == 0
    assert f"[{_CONFIG}] {_key} <- {_value}" in result.output

    # testing get
    result = cli_runner.invoke(get, ["--config", _CONFIG, _key])
    assert result.exit_code == 0
    assert f"[{_CONFIG}] {_key} = {_value}" in result.output


def test_config_show(cli_runner):
    _CONFIG = "CUSTOM"
    result = cli_runner.invoke(show)
    assert result.exit_code == 0
    assert "[DEFAULT]" in result.output

    result = cli_runner.invoke(show, ["--all"])
    assert "[DEFAULT]" in result.output
    assert "[CUSTOM]" in result.output


# NEGATIVE USE CASES
def test_config_get_set_missing_key(cli_runner):
    result = cli_runner.invoke(set, [])
    print(result.output)
    assert result.exit_code != 0
    assert "Error: Missing argument 'KEY'" in result.output

    runner = CliRunner()
    result = runner.invoke(get, [])
    print(result.output)
    assert result.exit_code != 0
    assert "Error: Missing argument 'KEY'" in result.output


def test_config_set_missing_value(cli_runner):
    result = cli_runner.invoke(set, ["key"])
    print(result.output)
    assert result.exit_code != 0
    assert "Error: Missing argument 'VALUE'" in result.output
